load("//node:toolchain.bzl", "NodeToolchainInfo")
load("@prelude//python:toolchain.bzl", "PythonToolchainInfo")

def _node_modules_impl(ctx: "context") -> [DefaultInfo.type]:
    out = ctx.actions.declare_output("node_modules")

    node_toolchain = ctx.attrs._node_toolchain[NodeToolchainInfo]

    cmd = cmd_args(
        ctx.attrs._python_toolchain[PythonToolchainInfo].interpreter,
        node_toolchain.build_node_modules[DefaultInfo].default_outputs,
        out.as_output(),
        "--package",
        ctx.attrs.package,
        "--package-lock",
        ctx.attrs.package_lock,
        "--bin-dir",
        node_toolchain.bin_dir[DefaultInfo].default_outputs,
    )

    for (name, value) in (ctx.attrs.extra_files or {}).items():
        cmd.add(cmd_args(value, format = "--extra-file=" + name + "={}"))

    if ctx.attrs.extra_args:
        cmd.add("--")
        cmd.add(ctx.attrs.extra_args)

    ctx.actions.run(cmd, category = "npm")

    return [DefaultInfo(default_output = out)]

node_modules = rule(
    impl = _node_modules_impl,
    attrs = {
        "package": attrs.source(),
        "package_lock": attrs.source(),
        "extra_files": attrs.option(
            attrs.dict(
                attrs.string(),
                attrs.source(allow_directory = True),
            ),
            default = None,
        ),
        "extra_args": attrs.option(
            attrs.list(attrs.arg()),
            default = None,
        ),
        "_node_toolchain": attrs.toolchain_dep(
            default = "toolchains//:node",
            providers = [NodeToolchainInfo],
        ),
        "_python_toolchain": attrs.toolchain_dep(
            default = "toolchains//:python",
            providers = [PythonToolchainInfo],
        ),
    },
)

def _npm_bin_impl(ctx: "context") -> [[DefaultInfo.type, RunInfo.type, TemplatePlaceholderInfo.type]]:
    bin_name = ctx.attrs.bin_name or ctx.attrs.name

    out = ctx.actions.declare_output(bin_name)

    node_toolchain = ctx.attrs._node_toolchain[NodeToolchainInfo]

    ctx.actions.run(
        [
            ctx.attrs._python_toolchain[PythonToolchainInfo].interpreter,
            node_toolchain.build_npm_bin[DefaultInfo].default_outputs,
            ctx.attrs.node_modules,
            bin_name,
            out.as_output(),
            "--bin-dir",
            node_toolchain.bin_dir[DefaultInfo].default_outputs,
        ],
        category = "build_npm_bin",
    )

    return [
        DefaultInfo(default_output = out),
        RunInfo(out),
        TemplatePlaceholderInfo(
            keyed_variables = {
                "exe": out,
            },
        ),
    ]

npm_bin = rule(
    impl = _npm_bin_impl,
    attrs = {
        "bin_name": attrs.option(attrs.string(), default = None),
        "node_modules": attrs.source(),
        "_node_toolchain": attrs.toolchain_dep(
            default = "toolchains//:node",
            providers = [NodeToolchainInfo],
        ),
        "_python_toolchain": attrs.toolchain_dep(
            default = "toolchains//:python",
            providers = [PythonToolchainInfo],
        ),
    },
)

def _npm_script_test_impl(ctx: "context") -> [[DefaultInfo.type, ExternalRunnerTestInfo.type]]:
    node_toolchain = ctx.attrs._node_toolchain[NodeToolchainInfo]

    cmd = [
        cmd_args(ctx.attrs._python_toolchain[PythonToolchainInfo].interpreter),
        cmd_args(node_toolchain.run_npm_script[DefaultInfo].default_outputs),
        "--node-modules",
        ctx.attrs.node_modules,
        "--bin-dir",
        cmd_args(node_toolchain.bin_dir[DefaultInfo].default_outputs),
    ]

    for (name, value) in (ctx.attrs.extra_files or {}).items():
        cmd.append(cmd_args(value, format = "--extra-file=" + name + "={}"))

    cmd.append(ctx.attrs.script or ctx.attrs.name)

    if ctx.attrs.args:
        cmd.append("--")
        cmd.extend(ctx.attrs.args)

    return [
        ExternalRunnerTestInfo(
            type = "npm",
            command = cmd,
            use_project_relative_paths = False,
        ),
        DefaultInfo(),
    ]

npm_script_test = rule(
    impl = _npm_script_test_impl,
    attrs = {
        "script": attrs.option(attrs.string(), default = None),
        "args": attrs.option(attrs.list(attrs.arg()), default = None),
        "node_modules": attrs.source(),
        "extra_files": attrs.option(
            attrs.dict(
                attrs.string(),
                attrs.source(allow_directory = True),
            ),
            default = None,
        ),
        "_node_toolchain": attrs.toolchain_dep(
            default = "toolchains//:node",
            providers = [NodeToolchainInfo],
        ),
        "_python_toolchain": attrs.toolchain_dep(
            default = "toolchains//:python",
            providers = [PythonToolchainInfo],
        ),
    },
)

def _prettier_diffs_impl(ctx: "context"):
    suggested_changes = []

    for src in ctx.attrs.srcs:
        suggestion = ctx.actions.declare_output("__suggested_changes__", src.short_path)
        ctx.actions.run(
            [
                "bash",
                "-xuo", "pipefail",
                "-c",
                """
                $prettier $SRC | diff -u${CONTEXT_SIZE} $SRC - > $OUT
                if test "$?" -eq 1; then
                  # we're fine with files being different; we'll catch that in
                  # the BXL script.
                  exit 0
                fi
                """,
            ],
            category = "prettier_diffs",
            identifier = src.short_path,
            env = {
                "prettier": ctx.attrs.prettier[RunInfo],
                "SRC": src,
                "OUT": suggestion.as_output(),
                "CONTEXT_SIZE": str(ctx.attrs.context_size),
            },
        )
        suggested_changes.append(suggestion)

    return [DefaultInfo(default_outputs = suggested_changes)]

prettier_diffs = rule(
    impl = _prettier_diffs_impl,
    attrs = {
        "srcs": attrs.list(attrs.source()),
        "context_size": attrs.int(default = 3),
        "prettier": attrs.dep(providers = [RunInfo]),
    }
)
